vim.g.jupynium_message_bloated = false

-- This number is not strictly accurate. We only use this to detect blockage (bloated)
vim.g.jupynium_num_pending_msgs = 0

if Jupynium_syncing_bufs ~= nil then
  for buf_id, _ in pairs(Jupynium_syncing_bufs) do
    -- This will clear autocmds if there are any
    vim.api.nvim_create_augroup(string.format("jupynium_buf_%d", buf_id), { clear = true })
  end
end

Jupynium_syncing_bufs = {} -- key = bufnr, value = 1

function Jupynium_reset_channel()
  vim.g.jupynium_channel_id = -1
  vim.g.jupynium_message_bloated = false
  vim.g.jupynium_num_pending_msgs = 0
  for buf_id, _ in pairs(Jupynium_syncing_bufs) do
    Jupynium_stop_sync(buf_id)
  end
end

local function get_line_count_all_buffers()
  local line_count = 0
  for buf_id, _ in pairs(Jupynium_syncing_bufs) do
    line_count = line_count + vim.api.nvim_buf_line_count(buf_id)
  end
  return line_count
end

get_line_count_all_buffers()

local function get_num_max_events()
  return math.max(vim.g.jupynium_num_max_msgs, get_line_count_all_buffers())
end

local function rpc(method, event, buf, ...)
  if vim.g.jupynium_channel_id ~= nil and vim.g.jupynium_channel_id > 0 then
    -- If bloated, wait until the messages are cleared
    -- Jupynium server should detect this and clear the messages
    -- And send grab_entire_buf
    if vim.g.jupynium_message_bloated then
      return
    end

    if vim.g.jupynium_num_pending_msgs < get_num_max_events() then
      vim.g.jupynium_num_pending_msgs = vim.g.jupynium_num_pending_msgs + 1
      local status, res = pcall(method, vim.g.jupynium_channel_id, event, buf, ...)
      if not status then
        print "Jupynium: RPC channel closed. Stop sending all notifications."
        Jupynium_reset_channel()
      else
        return res
      end
    else
      vim.g.jupynium_message_bloated = true
    end
  else
    Jupynium_reset_channel()
  end
end

function Jupynium_rpcnotify(event, buf, ...)
  -- check if it's already syncing
  if Jupynium_syncing_bufs[buf] == nil then
    return
  end
  rpc(vim.rpcnotify, event, buf, ...)
end

-- block until jupynium responds to the message
function Jupynium_rpcrequest(event, buf, ...)
  if Jupynium_syncing_bufs[buf] == nil then
    if event ~= "start_sync" then
      return
    end
  end

  local response = rpc(vim.rpcrequest, event, buf, ...)
  return response
end

function Jupynium_grab_entire_buffer(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot grab buffer without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  local entire_buf = vim.api.nvim_buf_get_lines(buf, 0, -1, false)
  Jupynium_rpcnotify("grab_entire_buf", buf, entire_buf)
end

function Jupynium_start_sync_cmd(args)
  local filename = args.args
  local buf = vim.api.nvim_get_current_buf()
  Jupynium_start_sync(buf, filename)
end

function Jupynium_start_sync(buf, filename)
  buf = buf or vim.api.nvim_get_current_buf()

  -- This will clear autocmds if there are any
  local augroup = vim.api.nvim_create_augroup(string.format("jupynium_buf_%d", buf), { clear = true })

  if Jupynium_syncing_bufs[buf] ~= nil then
    Jupynium_notify.error { "Already syncing this buffer.", ":JupyniumStopSync to stop." }
    return
  end

  local content = vim.api.nvim_buf_get_lines(buf, 0, -1, false)
  local response = Jupynium_rpcrequest("start_sync", buf, filename, content)
  if response ~= "OK" then
    Jupynium_notify.info { "Cancelling sync.." }
    return
  end

  Jupynium_syncing_bufs[buf] = 1

  vim.api.nvim_create_autocmd({ "CursorMoved" }, {
    buffer = buf,
    callback = function()
      local winid = vim.call("bufwinid", buf)
      local cursor_pos = vim.api.nvim_win_get_cursor(winid)
      local cursor_pos_row = cursor_pos[1] - 1
      local visual_start_row = vim.fn.getpos("v")[2] - 1
      Jupynium_rpcnotify("CursorMoved", buf, cursor_pos_row, visual_start_row)
    end,
    group = augroup,
  })

  vim.api.nvim_create_autocmd({ "CursorMovedI" }, {
    buffer = buf,
    callback = function()
      local winid = vim.call("bufwinid", buf)
      local cursor_pos = vim.api.nvim_win_get_cursor(winid)
      local cursor_pos_row = cursor_pos[1] - 1
      Jupynium_rpcnotify("CursorMovedI", buf, cursor_pos_row, cursor_pos_row)
    end,
    group = augroup,
  })

  vim.api.nvim_create_autocmd({ "ModeChanged" }, {
    buffer = buf,
    callback = function()
      local old_mode = vim.api.nvim_get_vvar("event")["old_mode"]
      local new_mode = vim.api.nvim_get_vvar("event")["new_mode"]
      if new_mode == "V" or new_mode == "v" or new_mode == "\x16" then
        local winid = vim.call("bufwinid", buf)
        local cursor_pos = vim.api.nvim_win_get_cursor(winid)
        local cursor_pos_row = cursor_pos[1] - 1
        local visual_start_row = vim.fn.getpos("v")[2] - 1
        Jupynium_rpcnotify("visual_enter", buf, cursor_pos_row, visual_start_row)
      elseif
        (old_mode == "v" or old_mode == "V" or old_mode == "\x16")
        and (new_mode ~= "v" and new_mode ~= "V" and new_mode ~= "\x16")
      then
        local winid = vim.call("bufwinid", buf)
        local cursor_pos = vim.api.nvim_win_get_cursor(winid)
        local cursor_pos_row = cursor_pos[1] - 1
        Jupynium_rpcnotify("visual_leave", buf, cursor_pos_row, cursor_pos_row)
      end
    end,
    group = augroup,
  })

  vim.api.nvim_create_autocmd({ "BufWritePre" }, {
    buffer = buf,
    callback = function()
      buf_filepath = vim.api.nvim_buf_get_name(buf)
      Jupynium_rpcnotify("BufWritePre", buf, buf_filepath)
    end,
    group = augroup,
  })

  vim.api.nvim_create_autocmd({ "BufWinLeave" }, {
    buffer = buf,
    callback = function()
      -- notify before detaching
      Jupynium_rpcnotify("BufWinLeave", buf)
      Jupynium_stop_sync(buf)
    end,
    group = augroup,
  })

  vim.api.nvim_buf_attach(buf, false, {
    on_lines = function(_, _, _, start_row, old_end_row, new_end_row, _)
      if Jupynium_syncing_bufs[buf] == nil then
        return
      end

      local lines = vim.api.nvim_buf_get_lines(buf, start_row, new_end_row, false)
      Jupynium_rpcnotify("on_lines", buf, lines, start_row, old_end_row, new_end_row)
    end,
  })
end

function Jupynium_stop_sync(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  Jupynium_rpcnotify("stop_sync", buf)
  -- This will clear autocmds if there are any
  vim.api.nvim_create_augroup(string.format("jupynium_buf_%d", buf), { clear = true })
  Jupynium_syncing_bufs[buf] = nil

  -- detach doesn't work. We just disable the on_lines callback by looking at Jupynium_syncing_bufs
  -- vim.api.nvim_buf_detach(buf)
end

function Jupynium_execute_selected_cells(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot execute cells without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  Jupynium_rpcnotify("execute_selected_cells", buf)
end

function Jupynium_toggle_selected_cells_outputs_scroll(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot toggle output cell without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  Jupynium_rpcnotify("toggle_selected_cells_outputs_scroll", buf)
end

function Jupynium_scroll_to_cell(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot scroll to cell without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  local winid = vim.call("bufwinid", buf)
  local cursor_pos = vim.api.nvim_win_get_cursor(winid)
  Jupynium_rpcnotify("scroll_to_cell", buf, cursor_pos[1] - 1)
end

function Jupynium_save_ipynb(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot save notebook without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  Jupynium_rpcnotify("save_ipynb", buf)
end

function Jupynium_download_ipynb(buf, output_name)
  buf = buf or vim.api.nvim_get_current_buf()
  -- get winnr from bufnr
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot download ipynb without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  buf_filepath = vim.api.nvim_buf_get_name(buf)
  if buf_filepath == "" then
    Jupynium_notify.error { [[Cannot download ipynb without having the filename for the buffer.]] }
    return
  end

  Jupynium_rpcnotify("download_ipynb", buf, buf_filepath, output_name)
end

function Jupynium_download_ipynb_cmd(args)
  local output_name = args.args
  Jupynium_download_ipynb(nil, output_name)
end

function Jupynium_auto_download_ipynb_toggle()
  vim.g.jupynium_autodownload_ipynb = 1 - vim.g.jupynium_auto_download_ipynb
  Jupynium_notify.info { "Auto download ipynb is now ", vim.g.jupynium_auto_download_ipynb == 1 and "on" or "off" }
end

function Jupynium_scroll_up(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot scroll notebook without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  local scroll_page = vim.g.jupynium_scroll_page or 0.5
  Jupynium_rpcnotify("scroll_ipynb", buf, -scroll_page)
end

function Jupynium_scroll_down(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot scroll notebook without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  local scroll_page = vim.g.jupynium_scroll_page or 0.5
  Jupynium_rpcnotify("scroll_ipynb", buf, scroll_page)
end

function Jupynium_autoscroll_toggle()
  vim.g.jupynium_autoscroll = 1 - vim.g.jupynium_autoscroll
  Jupynium_notify.info { "Autoscroll is now ", vim.g.jupynium_autoscroll == 1 and "on" or "off" }
end

function Jupynium_clear_selected_cells_outputs(buf)
  buf = buf or vim.api.nvim_get_current_buf()
  if Jupynium_syncing_bufs[buf] == nil then
    Jupynium_notify.error { [[Cannot clear outputs without synchronising.]], [[Run `:JupyniumStartSync`]] }
    return
  end

  Jupynium_rpcnotify("clear_selected_cells_outputs", buf)
end
